import vizibridge._vizibridge as rust_types
from vizibridge.kmers import Kmer, KmerType
from typing import Self, Iterator
from collections.abc import Mapping
from pathlib import Path
from dataclasses import dataclass


@dataclass
class KmerIndexEntry:
    kmer: KmerType | Kmer
    nid: int


kmer_indexes = [
    getattr(rust_types, a)
    for a in dir(rust_types)
    if a.startswith("KmerIndex") or a.startswith("LongKmerIndex")
]

KmerIndexType = kmer_indexes[0]
for t in kmer_indexes[1:]:
    KmerIndexType |= t

KmerIndexTypeMap = {KT.size(): KT for KT in kmer_indexes}


class KmerIndex(Mapping):
    __slots__ = ("__base_index", "__k")

    def __init__(self, path: Path, k: int):
        if not path.exists():
            raise IOError(path)
        self.__base_index = KmerIndexTypeMap[k](str(path))
        self.__k = k

    @property
    def k(self):
        return self.__k

    def __getitem__(self, kmer: KmerType | Kmer):
        if isinstance(kmer, KmerType):
            return self.__base_index[kmer]
        return self.__base_index[kmer.base_type]

    def __len__(self):
        return len(self.__base_index)

    def items_iter(self):
        return iter(self.__base_index)

    def __iter__(self):
        yield from (index_entry.kmer for index_entry in self.items_iter())

    @classmethod
    def build(
        cls,
        iterator: Iterator[KmerIndexEntry],
        index_path: Path,
        k: int,
        buffer_size=10**7,
    ) -> Self:
        BaseIndexType = KmerIndexTypeMap[k]
        if index_path.exists():
            raise IOError("path already exists")
        iterator = map(
            lambda e: KmerIndexEntry(
                kmer=getattr(e.kmer, "base_type", e.kmer), nid=e.nid
            ),
            iterator,
        )
        BaseIndexType.build(iterator, str(index_path), buffer_size=buffer_size)
        return cls(index_path, k)
